// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//

#include <winsock2.h>
#include <ws2tcpip.h>
#include <ws2spi.h>
#include "initguid.h"
#include "ndsupport.h"


typedef HRESULT
(*DLLGETCLASSOBJECT)(
    REFCLSID rclsid,
    REFIID rrid,
    LPVOID* ppv
    );

typedef HRESULT
(*DLLCANUNLOADNOW)(void);


HMODULE g_hProvider = NULL;
INDProvider* g_pIProvider = NULL;
DLLCANUNLOADNOW g_pfnDllCanUnloadNow = NULL;


static WCHAR* GetProviderPath( WSAPROTOCOL_INFOW* pProtocol )
{
    INT pathLen;
    INT ret, err;
    WCHAR* pPath;
    WCHAR* pPathEx;

    // Get the path length for the provider DLL.
    pathLen = 0;
    ret = WSCGetProviderPath( &pProtocol->ProviderId, NULL, &pathLen, &err );

    if( err != WSAEFAULT || pathLen == 0 )
        return NULL;

    pPath = (WCHAR*)HeapAlloc( GetProcessHeap(), 0, sizeof(WCHAR) * pathLen );
    if( pPath == NULL )
        return NULL;

    ret = WSCGetProviderPath( &pProtocol->ProviderId, pPath, &pathLen, &err );
    if( ret != 0 )
    {
        HeapFree( GetProcessHeap(), 0, pPath );
        return NULL;
    }

    pathLen = ExpandEnvironmentStringsW( pPath, NULL, 0 );
    if( pathLen == 0 )
    {
        HeapFree( GetProcessHeap(), 0, pPath );
        return NULL;
    }

    pPathEx = (WCHAR*)HeapAlloc( GetProcessHeap(), 0, sizeof(WCHAR) * pathLen );
    if( pPathEx == NULL )
    {
        HeapFree( GetProcessHeap(), 0, pPath );
        return NULL;
    }

    ret = ExpandEnvironmentStringsW( pPath, pPathEx, pathLen );

    // We don't need the un-expanded path anymore.
    HeapFree( GetProcessHeap(), 0, pPath );

    if( ret != pathLen )
    {
        HeapFree( GetProcessHeap(), 0, pPathEx );
        return NULL;
    }

    return pPathEx;
}


static HRESULT LoadProvider( WSAPROTOCOL_INFOW* pProtocol )
{
    WCHAR* pPath = GetProviderPath( pProtocol );
    if( pPath == NULL )
        return ND_UNSUCCESSFUL;

    g_hProvider = LoadLibraryW( pPath );
    if( g_hProvider == NULL )
        return __HRESULT_FROM_WIN32( GetLastError() );

    DLLGETCLASSOBJECT pfnDllGetClassObject =
        (DLLGETCLASSOBJECT)GetProcAddress( g_hProvider, "DllGetClassObject" );
    if( pfnDllGetClassObject == NULL )
    {
        return __HRESULT_FROM_WIN32( GetLastError() );
    }

    g_pfnDllCanUnloadNow =
        (DLLCANUNLOADNOW)GetProcAddress( g_hProvider, "DllCanUnloadNow" );
    if( g_pfnDllCanUnloadNow == NULL )
    {
        return __HRESULT_FROM_WIN32( GetLastError() );
    }

    IClassFactory* pClassFactory;
    HRESULT hr = pfnDllGetClassObject(
        pProtocol->ProviderId, IID_IClassFactory, (void**)&pClassFactory );
    if( FAILED(hr) )
        return hr;

    hr = pClassFactory->CreateInstance( NULL, IID_INDProvider, (void**)&g_pIProvider );

    // Now that we asked for the provider, we don't need the class factory.
    pClassFactory->Release();
    return hr;
}


static HRESULT Init()
{
    // Enumerate the provider catalog, find the first ND provider and load it.
    DWORD len = 0;
    INT err;
    INT ret = WSCEnumProtocols( NULL, NULL, &len, &err );
    if( ret != SOCKET_ERROR || err != WSAENOBUFS )
        return ND_INTERNAL_ERROR;

    WSAPROTOCOL_INFOW* pProtocols = (WSAPROTOCOL_INFOW*)HeapAlloc(
        GetProcessHeap(),
        0,
        len );
    if( pProtocols == NULL )
        return ND_NO_MEMORY;

    ret = WSCEnumProtocols( NULL, pProtocols, &len, &err );
    if( ret == SOCKET_ERROR )
        return ND_INTERNAL_ERROR;

    HRESULT hr;
    for( DWORD i = 0; i < len / sizeof(WSAPROTOCOL_INFOW); i++ )
    {
#define ServiceFlags1Flags (XP1_GUARANTEED_DELIVERY | XP1_GUARANTEED_ORDER | \
    XP1_MESSAGE_ORIENTED | XP1_CONNECT_DATA)

        if( (pProtocols[i].dwServiceFlags1 & ServiceFlags1Flags) !=
            ServiceFlags1Flags )
        {
            continue;
        }

        if( pProtocols[i].iAddressFamily != AF_INET &&
            pProtocols[i].iAddressFamily != AF_INET6 )
        {
            continue;
        }

        if( pProtocols[i].iSocketType != -1 )
            continue;

        if( pProtocols[i].iProtocol != 0 )
            continue;

        if( pProtocols[i].iProtocolMaxOffset != 0 )
            continue;

        hr = LoadProvider( &pProtocols[i] );
    }
    HeapFree( GetProcessHeap(), 0, pProtocols );

    return hr;
}


HRESULT ND_HELPER_API
NdStartup(
    VOID
    )
{
    int ret;
    WSADATA* pData = new WSADATA;
    if( !pData )
        return ND_NO_MEMORY;

    ret = WSAStartup( MAKEWORD(2, 2), pData );
    delete pData;
    if( ret != 0 )
        return ND_INTERNAL_ERROR;

    HRESULT hr = Init();
    if( FAILED( hr ) )
    {
        NdCleanup();
    }

    return hr;
}


HRESULT ND_HELPER_API
NdCleanup(
    VOID
    )
{
    if( g_pIProvider != NULL )
    {
        g_pIProvider->Release();
        g_pIProvider = NULL;
    }

    if( g_hProvider != NULL )
    {
        FreeLibrary( g_hProvider );
        g_hProvider = NULL;
    }

    WSACleanup();
    return S_OK;
}


VOID ND_HELPER_API
NdFlushProviders(
    VOID
    )
{
    return;
}


HRESULT ND_HELPER_API
NdQueryAddressList(
    __in DWORD Flags,
    __out_bcount_part_opt(*pBufferSize, *pBufferSize) SOCKET_ADDRESS_LIST* pAddressList,
    __inout SIZE_T* pBufferSize
    )
{
    UNREFERENCED_PARAMETER( Flags );

    if( g_pIProvider == NULL )
        return ND_NOT_INITIALIZED;

    return g_pIProvider->QueryAddressList( pAddressList, pBufferSize );
}


HRESULT ND_HELPER_API
NdResolveAddress(
    __in_bcount(RemoteAddressLength) const struct sockaddr* pRemoteAddress,
    __in SIZE_T RemoteAddressLength,
    __out_bcount(LocalAddressLength) struct sockaddr* pLocalAddress,
    __inout SIZE_T* pLocalAddressLength
    )
{
    SIZE_T len;

    // We store the original length so we can distinguish from different
    // errors that return WSAEFAULT.  We truncate lengths here, but that's OK.
    len = *pLocalAddressLength;
    // Create a socket for address changes.
    SOCKET s = WSASocketW( AF_INET, SOCK_STREAM, 0, NULL, 0, WSA_FLAG_OVERLAPPED );
    if( s == INVALID_SOCKET )
        return ND_INSUFFICIENT_RESOURCES;

    int ret = WSAIoctl( s, SIO_ROUTING_INTERFACE_QUERY,
        (VOID*)pRemoteAddress, (DWORD)RemoteAddressLength,
        pLocalAddress, (DWORD)len, (DWORD*)pLocalAddressLength, NULL, NULL );

    if( ret == SOCKET_ERROR )
    {
        switch( GetLastError() )
        {
        case WSAEFAULT:
            if( len < *pLocalAddressLength )
                return ND_BUFFER_OVERFLOW;
            // Fall through.
            __fallthrough;
        default:
            return ND_UNSUCCESSFUL;
        case WSAEINVAL:
        case WSAENETUNREACH:
        case WSAENETDOWN:
            return ND_INVALID_ADDRESS;
        }
    }

    return ND_SUCCESS;
}


HRESULT ND_HELPER_API
NdCheckAddress(
    __in_bcount(AddressLength) const struct sockaddr* pAddress,
    __in SIZE_T AddressLength
    )
{
    INDAdapter* pIAdapter;

    HRESULT hr = NdOpenIAdapter( pAddress, AddressLength, &pIAdapter );
    if( SUCCEEDED( hr ) )
    {
        pIAdapter->Release();
    }
    return hr;
}


HRESULT ND_HELPER_API
NdOpenIAdapter(
    __in_bcount(AddressLength) const struct sockaddr* pAddress,
    __in SIZE_T AddressLength,
    __deref_out INDAdapter** ppIAdapter
    )
{
    if( g_pIProvider == NULL )
        return ND_NOT_INITIALIZED;

    return g_pIProvider->OpenAdapter( pAddress, AddressLength, ppIAdapter );
}
